# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging

# Python 2/3 support
import sys
if sys.version_info[0] >= 3:    # pragma: no cover
    unicode = str

logger = logging.getLogger(__name__)


class InitDict(dict):
    """Base class of the data object classes to enforce required keys in the dict."""

    def __init__(self, **kwargs):
        """Check required keys."""
        if set(self.required_arguments).symmetric_difference(set(kwargs.keys())):
            raise AttributeError("required arguments: {}".format(self.required_arguments))
        super(InitDict, self).__init__(**kwargs)

    def __eq__(self, other):
        """Compare value of first key with specified unicode."""
        if isinstance(other, unicode):
            key_arg = self.required_arguments[0]
            return other == self[key_arg]
        return super(InitDict, self).__eq__(other)


class Volume(InitDict):
    """Data object representing a volume sortable by free size."""

    required_arguments = [
        "name", "state", "size_total", "size_used", "size_available", "files_used", "files_total", "snapable", "filer"
    ]

    def __lt__(self, other):
        """Use max() to get the volume with the biggest free size."""
        return self["size_available"] < other["size_available"]


class VolumeWithQuotaRatio(InitDict):
    """Data object representing a volume sortable by quota ratio."""

    required_arguments = [
        "name", "state", "size_total", "size_used", "size_available", "files_used", "files_total", "snapable",
        "quota_size", "quota_ratio", "filer"
    ]

    def __lt__(self, other):
        """Use min() to get the volume with the smallest ratio."""
        return self["quota_ratio"] < other["quota_ratio"]


class Snapshot(InitDict):
    """Data object representing a snapshot."""

    required_arguments = ["name"]


class QTree(InitDict):
    """Data object representing a qtree."""

    required_arguments = ["qtree", "status", "security_style"]


class Aggregate(InitDict):
    """Data object representing an aggregate."""

    required_arguments = ["id", "volume"]


class Quota(InitDict):
    """Data object representing a quota."""

    required_arguments = ["disk_limit", "file_limit", "threshold", "soft_disk_limit", "soft_file_limit"]


class QuotaReport(InitDict):
    """Data object representing a quota report."""

    required_arguments = [
        "disk_limit", "file_limit", "threshold", "soft_disk_limit", "soft_file_limit", "quota_target", "files_used",
        "disk_used", "tree"
    ]


class ACE(InitDict):
    """Data object representing an *access control entry*."""

    required_arguments = ["share_name", "permission", "user_or_group", "is_group", "user_group_type"]


class SnapmirrorStatus(InitDict):
    """Data object representing a snapmirror status."""

    required_arguments = [
        'source_location', 'destination_location', 'lag_time', 'last_transfer_from',
        'last_transfer_size', 'last_transfer_duration', 'last_transfer_type', 'status', 'transfer_progress',
        'mirror_timestamp', 'contents', 'state', 'base_snapshot', 'current_transfer_error', 'current_transfer_type',
        'inodes_replicated', 'replication_ops', 'break_failed_count', 'break_successful_count',
        'destination_volume', 'destination_volume_node', 'destination_vserver', 'destination_vserver_uuid',
        'exported_snapshot', 'exported_snapshot_timestamp', 'is_constituent', 'is_healthy',
        'last_transfer_end_timestamp', 'last_transfer_network_compression_ratio', 'max_transfer_rate', 'mirror_state',
        'newest_snapshot', 'newest_snapshot_timestamp', 'opmask', 'policy', 'policy_type',
        'relationship_control_plane', 'relationship_group_type', 'relationship_id', 'relationship_status',
        'relationship_type', 'resync_failed_count', 'resync_successful_count', 'source_volume', 'source_vserver',
        'source_vserver_uuid', 'total_transfer_bytes', 'total_transfer_time_secs', 'update_failed_count',
        'update_successful_count', 'vserver', 'snapmirror_status',
    ]


class SnapmirrorVolumeStatus(InitDict):
    """Data object representing a snapmirror volume status."""

    required_arguments = [
        "is_source", "is_destination", "is_transfer_in_progress", "is_transfer_broken"
    ]


class CifsShare(InitDict):
    """Data object representing a cifs share."""

    required_arguments = [
        "path", "share_name"
    ]


class SnapmirrorDestinationInfo(InitDict):
    """Data object representing a snapmirror destination info."""

    required_arguments = [
        "destination_location", "destination_volume", "destination_vserver", "is_constituent", "policy_type",
        "relationship_group_type", "relationship_id", "relationship_status", "relationship_type",
        "source_location", "source_volume", "source_volume_node", "source_vserver"
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals


def underline_to_dash(d):
    """Helper function to replace "_" to "-" in keys of specifed dictionary recursively.

    Netapp API uses "-" in XML parameters.

    :param d: dictionary of dictionaries or lists
    :type d: dict
    :return: new dictionary
    :rtype: dict
    """
    new = {}
    for k, v in d.items():
        if isinstance(v, dict):
            v = underline_to_dash(v)
        if isinstance(v, list):
            for i in v:
                v = [underline_to_dash(i)]
        new[k.replace('_', '-')] = v
    return new


def safe_get(d, key):
    """Helper function.

    :param d: dictionary
    :type d: dict
    :param key: key to retrieve from dict
    :type key: str
    :return: value of specified key if not None, otherwise empty dict
    """
    if key in d and d[key] is not None:
        return d[key]
    return {}



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

try:
    # py2
    from urlparse import urlparse
    from backports.functools_lru_cache import lru_cache
except ImportError:     # pragma: no cover
    # py3
    from urllib.parse import urlparse
    from functools import lru_cache

from cached_property import cached_property
import xml
import dicttoxml
import collections
import logging
from abc import ABCMeta, abstractmethod
# Python 2 and 3
from six import with_metaclass

from .http import NidhoggHttp
from .compatible import QuotaReport, Quota, QTree, VolumeWithQuotaRatio, Volume, SnapmirrorStatus
from .utils import underline_to_dash

# used the capture urllib3 wannings
logging.captureWarnings(True)
logger = logging.getLogger(__name__)

try:
    import pkg_resources
    version = pkg_resources.get_distribution('nidhogg').version
except:  # pragma: no cover
    version = 'development'


# after a quota resize wait max. 2 min til finished
# max wait time is QUOTA_RESIZE_WAIT_TIME * QUOTA_RESIZE_WAIT_CYCLES
QUOTA_RESIZE_WAIT_TIME = 6
QUOTA_RESIZE_WAIT_CYCLES = 20


class NidhoggException(Exception):
    """Exception wrapper."""

    pass


class Nidhogg(with_metaclass(ABCMeta)):
    """This is the base class for connecting to a NETAPP filer.

    It provides functions that have 7-mode filers and cluster-mode filers in common.

    Subclasses:

    * :class:`~nidhogg.sevenmode.SevenMode`
    * :class:`~nidhogg.clustermode.ClusterMode`
    """

    def __init__(self, url, username, password, major, minor, verify, http=NidhoggHttp):
        """Init conncetion to filer."""
        self.url = url
        self.major = major
        self.minor = minor
        self.http = http(url, username, password, verify)
        self.xmlns = "http://www.netapp.com/filer/admin"
        self.nmsdk_version = version
        self.nmsdk_language = "python"
        self.nmsdk_app = "Nidhogg"

    def __getattr__(self, api):
        """Try to invoke unimplemented API calls directly."""
        def _api_wrapper(**kwargs):
            return self._do(api, **kwargs)
        return _api_wrapper

    def _do(self, api, **kwargs):
        """Invoke wrapper, returns a xmldict."""
        # replace _ -> -
        params = underline_to_dash(kwargs)
        req = self._create_request(**{api.replace("_", "-"): params})
        logger.debug("request: {0}".format(req))
        r = self.http.invoke_request(req)
        logger.debug("response: {0}".format(r))
        try:
            self.xmldict = self.http.parse_xml_reply(r)
        except xml.parsers.expat.ExpatError:
            logger.exception("exception on {}".format(self.vserver))
            raise NidhoggException(r + " (host: {})".format(self.vserver))

        if self.xmldict["netapp"]["results"]['@status'] == "failed":
            logger.error("exception on {}".format(self.vserver))
            logger.error(self.xmldict["netapp"]["results"]['@reason'])
            logger.error("{0} failed with params {1}".format(api, params))
            raise NidhoggException(self.xmldict["netapp"]["results"]['@reason'] + " (host: {})".format(self.vserver))
        return self.xmldict

    def _create_request(self, **kwargs):
        xml_request = dicttoxml.dicttoxml(collections.OrderedDict(**kwargs), root=False, attr_type=False)
        return """
            <?xml version='1.0' encoding='utf-8'?>
            <netapp version='{0.major}.{0.minor}'
                    xmlns='{0.xmlns}'
                    nmsdk_version='{0.nmsdk_version}'
                    nmsdk_language='{0.nmsdk_language}'
                    nmsdk_app='{0.nmsdk_app}'>
                {1}
            </netapp>""".strip().replace("            ", "").format(self, xml_request.decode("utf-8"))

    def _item_to_quota(self, item):
        """Convert to byte (API returns sizes in kbyte) to be consistent with other sizes (i.e. volume info)."""
        return Quota(
            disk_limit=float(item['disk-limit']) * 1024 if item['disk-limit'].isdigit() else -1,
            soft_disk_limit=float(item['soft-disk-limit']) * 1024 if item['soft-disk-limit'].isdigit() else -1,
            threshold=float(item['threshold']) * 1024 if item['threshold'].isdigit() else -1,
            file_limit=float(item['file-limit']) if item['file-limit'].isdigit() else -1,
            soft_file_limit=float(item['soft-file-limit']) if item['soft-file-limit'].isdigit() else -1,
        )

    def _item_to_quota_report(self, item):
        return QuotaReport(
            disk_used=float(item['disk-used']) * 1024 if item['disk-used'].isdigit() else -1,
            disk_limit=float(item['disk-limit']) * 1024 if item['disk-limit'].isdigit() else -1,
            soft_disk_limit=float(item['soft-disk-limit']) * 1024 if item['soft-disk-limit'].isdigit() else -1,
            threshold=float(item['threshold']) * 1024 if item['threshold'].isdigit() else -1,
            files_used=float(item['files-used']) if item['files-used'].isdigit() else -1,
            file_limit=float(item['file-limit']) if item['file-limit'].isdigit() else -1,
            soft_file_limit=float(item['soft-file-limit']) if item['soft-file-limit'].isdigit() else -1,
            quota_target=item['quota-target'],
            tree=item['tree'],
        )

    def _item_to_qtree(self, item):
        return QTree(
            qtree=item['qtree'],
            status=item['status'],
            security_style=item['security-style'],
        )

    def _item_to_snapmirrorstatus(self, item):
        return SnapmirrorStatus(
            # 7mode and cluster mode
            source_location=item["source-location"],
            destination_location=item["destination-location"],
            lag_time=item["lag-time"],
            last_transfer_from=item["last-transfer-from"],
            last_transfer_size=item["last-transfer-size"],
            last_transfer_duration=item["last-transfer-duration"],
            last_transfer_type=item['last-transfer-type'] if 'last-transfer-type' in item else None,
            # sevenmode
            status=item["status"] if 'status' in item else None,
            transfer_progress=item["transfer-progress"] if 'transfer-progress' in item else None,
            mirror_timestamp=item["mirror-timestamp"] if 'mirror-timestamp' in item else None,
            contents=item["contents"] if 'contents' in item else None,
            state=item["state"] if 'state' in item else None,
            # sevenmode optional
            base_snapshot=item['base-snapshot'] if 'base-snapshot' in item else None,
            current_transfer_error=item['current-transfer-error'] if 'current-transfer-error' in item else None,
            current_transfer_type=item['current-transfer-type'] if 'current-transfer-type' in item else None,
            inodes_replicated=item['inodes-replicated'] if 'inodes-replicated' in item else None,
            replication_ops=item['replication-ops'] if 'replication-ops' in item else None,
            # cluster mode
            break_failed_count=item['break-failed-count'] if 'break-failed-count' in item else None,
            break_successful_count=item['break-successful-count'] if 'break-successful-count' in item else None,
            destination_volume=item['destination-volume'] if 'destination-volume' in item else None,
            destination_volume_node=item['destination-volume-node'] if 'destination-volume-node' in item else None,
            destination_vserver=item['destination-vserver'] if 'destination-vserver' in item else None,
            destination_vserver_uuid=item['destination-vserver-uuid'] if 'destination-vserver-uuid' in item else None,
            exported_snapshot=item['exported-snapshot'] if 'exported-snapshot' in item else None,
            exported_snapshot_timestamp=item['exported-snapshot-timestamp'] if 'exported-snapshot-timestamp' in item else None,
            is_constituent=item['is-constituent'] if 'is-constituent' in item else None,
            is_healthy=item['is-healthy'] if 'is-healthy' in item else None,
            last_transfer_end_timestamp=item['last-transfer-end-timestamp'] if 'last-transfer-end-timestamp' in item else None,
            last_transfer_network_compression_ratio=item['last-transfer-network-compression-ratio'] if 'last-transfer-network-compression-ratio' in item else None,
            max_transfer_rate=item['max-transfer-rate'] if 'max-transfer-rate' in item else None,
            mirror_state=item['mirror-state'] if 'mirror-state' in item else None,
            newest_snapshot=item['newest-snapshot'] if 'newest-snapshot' in item else None,
            newest_snapshot_timestamp=item['newest-snapshot-timestamp'] if 'newest-snapshot-timestamp' in item else None,
            opmask=item['opmask'] if 'opmask' in item else None,
            policy=item['policy'] if 'policy' in item else None,
            policy_type=item['policy-type'] if 'policy-type' in item else None,
            relationship_control_plane=item['relationship-control-plane'] if 'relationship-control-plane' in item else None,
            relationship_group_type=item['relationship-group-type'] if 'relationship-group-type' in item else None,
            relationship_id=item['relationship-id'] if 'relationship-id' in item else None,
            relationship_status=item['relationship-status'] if 'relationship-status' in item else None,
            relationship_type=item['relationship-type'] if 'relationship-type' in item else None,
            resync_failed_count=item['resync-failed-count'] if 'resync-failed-count' in item else None,
            resync_successful_count=item['resync-successful-count'] if 'resync-successful-count' in item else None,
            source_volume=item['source-volume'] if 'source-volume' in item else None,
            source_vserver=item['source-vserver'] if 'source-vserver' in item else None,
            source_vserver_uuid=item['source-vserver-uuid'] if 'source-vserver-uuid' in item else None,
            total_transfer_bytes=item['total-transfer-bytes'] if 'total-transfer-bytes' in item else None,
            total_transfer_time_secs=item['total-transfer-time-secs'] if 'total-transfer-time-secs' in item else None,
            update_failed_count=item['update-failed-count'] if 'update-failed-count' in item else None,
            update_successful_count=item['update-successful-count'] if 'update-successful-count' in item else None,
            vserver=item['vserver'] if 'vserver' in item else None,
            # helper state, mapping relationship-status (cluster mode) and status (7mode) to snapmirror-status
            snapmirror_status=item['relationship-status'].lower() if self.clustered else item["status"].lower(),
        )

    #
    # cached PROPERTIES
    #
    @cached_property
    def vserver(self):
        """Hostname of the connected filer.

        :return: hostname
        :rtype: str
        """
        return str(urlparse(self.url).hostname).split(".")[0]

    @cached_property
    def vserver_fqdn(self):
        """FQDN of the connected filer.

        :return: FQDN of the filer
        :rtype: str
        """
        return str(urlparse(self.url).hostname)

    @cached_property
    def clustered(self):
        """True if the filer is a cluster-mode filer, false otherwise.

        :rtype: boolean
        """
        results = self.system_get_version()['netapp']['results']
        return 'is-clustered' in results and results['is-clustered'] == "true"

    @cached_property
    def ontapi_version(self):
        """ONTAPI version of the connected filer.

        :return: ontapi version
        :rtype: str
        """
        # get result
        result = self.system_get_ontapi_version()['netapp']['results']
        # split afterwards, does not work the short way ?:|
        return "{0}.{1}".format(result['major-version'], result['minor-version'])

    @cached_property
    def has_forcegroup(self):
        """Check if this cifs share feature is available.

        :return: true, if feature is available
        :rtype: bool
        """
        # sevenmode supports force group
        if not self.clustered:
            return True
        # clustermode with ontapi 1.30 and onwards
        result = self.system_get_ontapi_version()['netapp']['results']
        return int(result['major-version']) >= 1 and int(result['minor-version']) >= 30

    @cached_property
    def apis(self):
        """List of API commands available with the current credentials.

        :return: list of API commands
        :rtype: list of str or empty list
        """
        try:
            return [
                item["name"]
                for item in self.system_api_list()["netapp"]["results"]["apis"]["system-api-info"]
            ]
        except NidhoggException:
            return []

    #
    # COMMON API FUNCTIONS
    #
    def list_snapable_volumes(self):
        """Return a list of *snapable* volumes.

        That means, ignore volumes that are used as a snapmirror destination.

        :return: list of snapable volumes
        :rtype: list of :class:`~nidhogg.compatible.Volume`
        :raises NidhoggException: if an error occurs
        """
        vols = []
        for vol in self.list_volumes():
            if vol['state'] == 'online' and vol['snapable']:
                vols.append(vol)
        return vols

    def delete_snapshot(self, volume, name):
        """Delete a snapshot.

        :param volume: name of the volume
        :type volume: str
        :param name: name of the snapshot
        :type name: str
        :raises NidhoggException: if an error occurs
        """
        return self.snapshot_delete(volume=volume, snapshot=name)

    def create_qtree(self, volume, qtree, mode="007"):
        """Create a qtree on the specified volume.

        :param volume: name of the volume
        :type volume: str
        :param qtree: name of the qtree to be created
        :type name: str
        :param mode: initial file system permissions of the qtree
        :type mode: str
        :raises NidhoggException: if an error occurs
        """
        self.qtree_create(volume=volume, qtree=qtree, mode=mode)

    def delete_qtree(self, volume, qtree, force=False):
        """Delete a qtree on the specified volume.

        :param volume: name of the volume
        :type volume: str
        :param qtree: name of the qtree to be deleted
        :type qtree: str
        :param force: force deletion if true
        :type force: bool
        :raises NidhoggException: if an error occurs
        """
        self.qtree_delete(force=str(bool(force)), qtree="/vol/{0}/{1}".format(volume, qtree))

    def delete_cifs_share(self, share_name):
        """Delete the share with the given name.

        :param share_name: name of the share
        :type share_name: str
        :raises NidhoggException: if an error occurs
        """
        self.cifs_share_delete(share_name=share_name)

    def exists_qtree(self, volume, qtree):
        """Check if a qtree exits.

        :param volume: name of the volume
        :type volume: str
        :param qtree: name of the qtree
        :type qtree: str
        :return: true, if the qtree exists
        :rtype: bool
        :raises NidhoggException: if an error occurs
        """
        return len([qt for qt in self.list_qtrees(volume) if qt['qtree'] == qtree]) > 0

    #
    # cached API functions
    #
    @lru_cache(maxsize=100)
    def get_allocated_quota_size(self, volume):
        """Return the sum of all quotas of the specified volume.

        :param volume: name of the volume
        :type volume: str
        :return: sum of all qtree quotas on this volume in byte
        :rtype: int
        :raises NidhoggException: if an error occurs
        """
        # only use those where a tree is specified
        return sum(quota['disk_limit'] for quota in self.list_quotas(volume) if quota['tree'])

    @lru_cache(maxsize=100)
    def get_allocated_quota_ratio(self, volume, volume_size_total=None):
        """Return the ratio *allocated quota size / volume size*.

        :param volume: name of the volume
        :type volume: str
        :param volume_size_total: if specified we have the total size already from a previous API call
        :type volume_size_total: int
        :return: ratio *allocated quota size / volume size*
        :rtype: int
        :raises NidhoggException: if an error occurs
        """
        # shortcut: if specified we have the size already from a previous api call
        if volume_size_total:
            return self.get_allocated_quota_size(volume) / volume_size_total
        return self.get_allocated_quota_size(volume) / self.volume_info(volume)["size_total"]

    def get_volumes_with_quota_info(self, filter_volume_names=[]):
        """Return a list of snapable volumes of type :class:`~nidhogg.compatible.VolumeWithQuotaRatio`.

        :param filter_volume_names: consider only volumes that are in this list
        :type filter_volume_names: list of str
        :return: list of project home volumes
        :rtype: list of :class:`~nidhogg.compatible.VolumeWithQuotaRatio`
        :raises NidhoggException: if an error occurs
        """
        volumes = []
        # get all volumes with type "rw"
        for v in self.list_snapable_volumes():
            # return only volumes with matching volume names if specified
            if filter_volume_names:
                if v["name"] not in filter_volume_names:
                    logger.debug("filer: {0}, skipped volume '{1}' because not in filter".format(
                        self.vserver_fqdn,
                        v["name"])
                    )
                    continue
            project_volume = VolumeWithQuotaRatio(
                quota_size=self.get_allocated_quota_size(volume=v["name"]),
                quota_ratio=self.get_allocated_quota_ratio(
                    volume=v["name"],
                    volume_size_total=v["size_total"]),
                **v
            )
            volumes.append(project_volume)
        return volumes

    def get_volumes(self, filter_volume_names=[]):
        """Return a list of snapable volumes of type :class:`~nidhogg.compatible.Volume`.

        :param filter_volume_names: consider only volumes that are in this list
        :type filter_volume_names: list of str
        :return: list of user home volumes
        :rtype: list of :class:`~nidhogg.compatible.Volume`
        :raises NidhoggException: if an error occurs
        """
        volumes = []
        # get all volumes with type "rw"
        for v in self.list_snapable_volumes():
            # return only volumes with matching volume names
            if filter_volume_names:
                if v["name"] not in filter_volume_names:
                    logger.debug("filer: {0}, skipped volume '{1}' because not in filter".format(
                        self.vserver_fqdn,
                        v["name"])
                    )
                    continue
            volumes.append(Volume(**v))
        return volumes

    #
    # API FUNCTIONS implemented in subclasses
    #
    @abstractmethod
    def list_qtrees(self, *args, **kwargs):
        """See sub classes.

        * Go to :py:meth:`~.SevenMode.list_qtrees` (SevenMode)
        * Go to :py:meth:`~.ClusterMode.list_qtrees` (ClusterMode)
        """
        pass    # pragma: no cover

    @abstractmethod
    def list_volumes(self, *args, **kwargs):
        """See sub classes.

        * Go to :py:meth:`~.SevenMode.list_volumes` (SevenMode)
        * Go to :py:meth:`~.ClusterMode.list_volumes` (ClusterMode)
        """
        pass    # pragma: no cover

    @abstractmethod
    def volume_info(self, *args, **kwargs):
        """See sub classes.

        * Go to :py:meth:`~.SevenMode.volume_info` (SevenMode)
        * Go to :py:meth:`~.ClusterMode.volume_info` (ClusterMode)
        """
        pass    # pragma: no cover

    @abstractmethod
    def list_snapshots(self, *args, **kwargs):
        """See sub classes.

        * Go to :py:meth:`~.SevenMode.list_snapshots` (SevenMode)
        * Go to :py:meth:`~.ClusterMode.list_snapshots` (ClusterMode)
        """
        pass    # pragma: no cover

    @abstractmethod
    def get_quota(self, *args, **kwargs):
        """See sub classes.

        * Go to :py:meth:`~.SevenMode.get_quota` (SevenMode)
        * Go to :py:meth:`~.ClusterMode.get_quota` (ClusterMode)
        """
        pass    # pragma: no cover

    @abstractmethod
    def list_quotas(self, *args, **kwargs):
        """See sub classes.

        * Go to :py:meth:`~.SevenMode.list_quotas` (SevenMode)
        * Go to :py:meth:`~.ClusterMode.list_quotas` (ClusterMode)
        """
        pass    # pragma: no cover

    @abstractmethod
    def list_cifs_shares(self, *args, **kwargs):
        """See sub classes.

        * Go to :py:meth:`~.SevenMode.list_cifs_shares` (SevenMode)
        * Go to :py:meth:`~.ClusterMode.list_cifs_shares` (ClusterMode)
        """
        pass    # pragma: no cover

    @abstractmethod
    def create_cifs_share(self, *args, **kwargs):
        """See sub classes.

        * Go to :py:meth:`~.SevenMode.create_cifs_share` (SevenMode)
        * Go to :py:meth:`~.ClusterMode.create_cifs_share` (ClusterMode)
        """
        pass    # pragma: no cover

    @abstractmethod
    def set_cifs_acl(self, *args, **kwargs):
        """See sub classes.

        * Go to :py:meth:`~.SevenMode.set_cifs_acl` (SevenMode)
        * Go to :py:meth:`~.ClusterMode.set_cifs_acl` (ClusterMode)
        """
        pass    # pragma: no cover

    @abstractmethod
    def list_cifs_acls(self, *args, **kwargs):
        """See sub classes.

        * Go to :py:meth:`~.SevenMode.list_cifs_acls` (SevenMode)
        * Go to :py:meth:`~.ClusterMode.list_cifs_acls` (ClusterMode)
        """
        pass    # pragma: no cover

    @abstractmethod
    def delete_cifs_acl(self, *args, **kwargs):
        """See sub classes.

        * Go to :py:meth:`~.SevenMode.delete_cifs_acl` (SevenMode)
        * Go to :py:meth:`~.ClusterMode.delete_cifs_acl` (ClusterMode)
        """
        pass    # pragma: no cover

    @abstractmethod
    def delete_cifs_acls(self, *args, **kwargs):
        """See sub classes.

        * Go to :py:meth:`~.SevenMode.delete_cifs_acls` (SevenMode)
        * Go to :py:meth:`~.ClusterMode.delete_cifs_acls` (ClusterMode)
        """
        pass    # pragma: no cover

    @abstractmethod
    def set_quota(self, *args, **kwargs):
        """See sub classes.

        * Go to :py:meth:`~.SevenMode.set_quota` (SevenMode)
        * Go to :py:meth:`~.ClusterMode.set_quota` (ClusterMode)
        """
        pass    # pragma: no cover

    @abstractmethod
    def delete_quota(self, *args, **kwargs):
        """See sub classes.

        * Go to :py:meth:`~.SevenMode.delete_quota` (SevenMode)
        * Go to :py:meth:`~.ClusterMode.delete_quota` (ClusterMode)
        """
        pass    # pragma: no cover

    @abstractmethod
    def update_snapmirror(self, *args, **kwargs):
        """See sub classes.

        * Go to :py:meth:`~.SevenMode.update_snapmirror` (SevenMode)
        * Go to :py:meth:`~.ClusterMode.update_snapmirror` (ClusterMode)
        """
        pass    # pragma: no cover

    @abstractmethod
    def update_snapmirror_with_snapshot(self, *args, **kwargs):
        """See sub classes.

        * Go to :py:meth:`~.SevenMode.update_snapmirror_with_snapshot` (SevenMode)
        * Go to :py:meth:`~.ClusterMode.update_snapmirror_with_snapshot` (ClusterMode)
        """
        pass    # pragma: no cover

    @abstractmethod
    def get_snapmirror_status(self, *args, **kwargs):
        """See sub classes.

        * Go to :py:meth:`~.SevenMode.get_snapmirror_status` (SevenMode)
        * Go to :py:meth:`~.ClusterMode.get_snapmirror_status` (ClusterMode)
        """
        pass    # pragma: no cover

    @abstractmethod
    def get_snapmirror_volume_status(self, *args, **kwargs):
        """See sub classes.

        * Go to :py:meth:`~.SevenMode.get_snapmirror_volume_status` (SevenMode)
        * Go to :py:meth:`~.ClusterMode.get_snapmirror_volume_status` (ClusterMode)
        """
        pass    # pragma: no cover

    @abstractmethod
    def create_snapshot(self, *args, **kwargs):
        """See sub classes.

        * Go to :py:meth:`~.SevenMode.create_snapshot` (SevenMode)
        * Go to :py:meth:`~.ClusterMode.create_snapshot` (ClusterMode)
        """
        pass    # pragma: no cover

    @abstractmethod
    def list_snapmirror_destinations(self, *args, **kwargs):
        """See sub classes.

        * Go to :py:meth:`~.SevenMode.list_snapmirror_destinations` (SevenMode)
        * Go to :py:meth:`~.ClusterMode.list_snapmirror_destinations` (ClusterMode)
        """
        pass    # pragma: no cover



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

try:
    # py2
    from backports.functools_lru_cache import lru_cache
except ImportError:     # pragma: no cover
    # py3
    from functools import lru_cache

import nidhogg.core     # this style needed for patching

from .compatible import ACE, CifsShare, SnapmirrorDestinationInfo, Snapshot, Volume
from .core import Nidhogg, NidhoggException
from time import sleep

import logging
logger = logging.getLogger(__name__)


#: maximum records that can be retrieved via NETAPP API
MAX_RECORDS = 2 ** 16


class ClusterMode(Nidhogg):
    """This class implements cluster-mode filer specific API calls."""

    ACL_FULL_CONTROL = "full_control"       #: ACL permission constant for full control
    ACL_READ = "read"                       #: ACL permission constant for read access
    ACL_CHANGE = "change"                   #: ACL permission constant for write access
    ACL_NO_ACCESS = "no_access"             #: ACL permission constant for denying access
    ACL_PERMISSIONS = [
        ACL_FULL_CONTROL,
        ACL_READ,
        ACL_CHANGE,
        ACL_NO_ACCESS
    ]                                       #: list of all permission constants

    def _item_to_volume(self, item):
        return Volume(
            name=item['volume-id-attributes']['name'],
            # RW for read-write, DP for data-protection, DC for data-cache, LS for load-sharing
            snapable=item['volume-id-attributes']['type'] == "rw",
            state=item['volume-state-attributes']['state'] if 'volume-state-attributes' in item and 'state' in item['volume-state-attributes'] else None,
            size_total=float(item['volume-space-attributes']['size-total']) if 'volume-space-attributes' in item and 'size-total' in item['volume-space-attributes'] else None,
            size_used=float(item['volume-space-attributes']['size-used']) if 'volume-space-attributes' in item and 'size-used' in item['volume-space-attributes'] else None,
            size_available=float(item['volume-space-attributes']['size-available']) if 'volume-space-attributes' in item and 'size-available' in item['volume-space-attributes'] else None,
            files_used=float(item['volume-inode-attributes']['files-used']) if 'volume-inode-attributes' in item and 'files-used' in item['volume-inode-attributes'] else None,
            files_total=float(item['volume-inode-attributes']['files-total']) if 'volume-inode-attributes' in item and 'files-total' in item['volume-inode-attributes'] else None,
            filer=self.vserver_fqdn,
        )

    def _item_to_ace(self, item):
        return ACE(
            share_name=item['share'],
            user_or_group=item['user-or-group'],
            permission=item['permission'],
            is_group=None,      # not used
            user_group_type=item['user-group-type'] if "user-group-type" in item else None
        )

    def _item_to_snapmirrordestinationinfo(self, item):
        return SnapmirrorDestinationInfo(
            destination_location=item["destination-location"],
            destination_volume=item['destination-volume'],
            destination_vserver=item['destination-vserver'],
            is_constituent=item['is-constituent'],
            policy_type=item['policy-type'],
            relationship_group_type=item['relationship-group-type'],
            relationship_id=item['relationship-id'],
            relationship_status=item['relationship-status'],
            relationship_type=item['relationship-type'],
            source_location=item["source-location"],
            source_volume=item['source-volume'],
            source_volume_node=item['source-volume-node'],
            source_vserver=item['source-vserver']
        )

    #
    # API FUNCTIONS
    #
    def list_qtrees(self, volume, max_records=MAX_RECORDS):
        """Return a list of qtrees of type :class:`~nidhogg.compatible.QTree`.

        :param volume: name of the volume
        :type volume: str
        :param max_records: limit returned records
        :type max_records: int
        :return: list of qtrees
        :rtype: list of :class:`~nidhogg.compatible.QTree` or empty list
        :raises NidhoggException: if an error occurs
        """
        opts = dict(
            query=dict(
                qtree_info=dict(
                    volume=volume
                )
            ),
            max_records=max_records
        )
        results = self.qtree_list_iter(**opts)["netapp"]["results"]
        if int(results["num-records"]) > 1:
            return [
                self._item_to_qtree(item)
                for item in results["attributes-list"]["qtree-info"]
            ]
        elif int(results["num-records"]) == 1:
            return [
                self._item_to_qtree(results["attributes-list"]["qtree-info"])
            ]
        logger.warn("list_qtrees: no entries found")
        return []

    @lru_cache(maxsize=100)
    def list_volumes(self, max_records=MAX_RECORDS):
        """Return a list of volumes of type :class:`~nidhogg.compatible.Volume`.

        :param max_records: limit returned records
        :type max_records: int
        :return: list of volumes
        :rtype: list of :class:`~nidhogg.compatible.Volume` or empty list
        :raises NidhoggException: if an error occurs
        """
        opts = dict(
            max_records=max_records
        )
        results = self.volume_get_iter(**opts)["netapp"]["results"]
        if int(results["num-records"]) > 1:
            return [
                self._item_to_volume(item)
                for item in results["attributes-list"]["volume-attributes"]
            ]
        elif int(results["num-records"]) == 1:
            return [
                self._item_to_volume(results["attributes-list"]["volume-attributes"])
            ]
        logger.warn("list_volumes: no entries found")
        return []

    @lru_cache(maxsize=100)
    def volume_info(self, volume):
        """Return basic information about the volume.

        :param volume: name of the volume
        :type volume: str
        :return: volume
        :rtype: :class:`~nidhogg.compatible.Volume`
        :raises NidhoggException: if an error occurs
        """
        opts = dict(
            query=dict(
                volume_id_attributes=dict(
                    name=volume
                )
            )
        )
        return self._item_to_volume(
            self.volume_get_iter(**opts)["netapp"]["results"]["attributes-list"]["volume-attributes"])

    def list_snapshots(self, target_name, max_records=MAX_RECORDS):
        """Return list of snapshots for given volume.

        :param target_name: name of the volume
        :type target_name: str
        :param max_records: limit returned records
        :type max_records: int
        :return: list of snapshots
        :rtype: list of :class:`~nidhogg.compatible.Snapshot` or empty list
        :raises NidhoggException: if an error occurs
        """
        opts = dict(
            query=dict(
                snapshot_info=dict(
                    volume=target_name
                )
            ),
            max_records=max_records
        )
        results = self.snapshot_get_iter(**opts)["netapp"]["results"]
        if int(results["num-records"]) > 1:
            return [
                Snapshot(name=item['name'])
                for item in results["attributes-list"]["snapshot-info"]
            ]
        elif int(results["num-records"]) == 1:
            return [
                Snapshot(name=results["attributes-list"]["snapshot-info"]['name'])
            ]
        logger.warn("list_snapshots: no entries found")
        return []

    def get_quota(self, volume, qtree, max_records=MAX_RECORDS):
        """Return the quota of the specified qtree on the given volume.

        :param volume: name of the volume
        :type volume: str
        :param qtree: name of the qtree
        :type qtree: str
        :param max_records: limit returned records
        :type max_records: int
        :return: quota
        :rtype: :class:`~nidhogg.compatible.Quota` or empty dict
        :raises NidhoggException: if an error occurs
        """
        opts = dict(
            query=dict(
                quota_entry=dict(
                    quota_target="/vol/{0}/{1}".format(volume, qtree)
                )
            ),
            max_records=max_records
        )
        results = self.quota_list_entries_iter(**opts)["netapp"]["results"]
        if int(results["num-records"]) == 1:
            return self._item_to_quota(results['attributes-list']['quota-entry'])
        logger.warn("get_quota: no entries found")
        return {}

    def list_quotas(self, volume, max_records=MAX_RECORDS):
        """Return a list of quota reports of the specified volume.

        :param volume: name of the volume
        :type volume: str
        :param max_records: limit returned records
        :type max_records: int
        :return: list of quota reports
        :rtype: :class:`~nidhogg.compatible.QuotaReport` or empty list
        :raises NidhoggException: if an error occurs
        """
        opts = dict(
            query=dict(
                quota=dict(
                    volume=volume
                )
            ),
            max_records=max_records
        )
        results = self.quota_report_iter(**opts)["netapp"]["results"]
        if int(results["num-records"]) > 1:
            return [
                self._item_to_quota_report(item)
                for item in results['attributes-list']['quota']
            ]
        elif int(results["num-records"]) == 1:
            return [
                self._item_to_quota_report(results['attributes-list']['quota'])
            ]
        logger.warn("list_quotas: no entries found")
        return []

    def list_cifs_shares(self):
        """List all cifs shares.

        :return: list of cifs shares
        :rtype: list of :class:`~nidhogg.compatible.CifsShare` or empty list
        :raises NidhoggException: if an error occurs
        """
        opts = dict(
            max_records=2 ** 32 - 1
        )

        results = self.cifs_share_get_iter(**opts)['netapp']['results']

        if int(results['num-records']) > 1:
            return [
                CifsShare(path=item['path'], share_name=item['share-name'])
                for item in results['attributes-list']['cifs-share']
            ]
        elif int(results['num-records']) == 1:
            return [
                CifsShare(path=results['attributes-list']['cifs-share']['path'], share_name=results['attributes-list']['cifs-share']['share-name'])
            ]
        logger.warning("list_cifs_shares: cifs shares found")
        return []

    def create_cifs_share(self, volume, qtree, share_name, group_name=None, comment=None, umask="007", vscan_fileop_profile="standard"):
        """Create a cifs share.

        :param volume: name of the volume
        :type volume: str
        :param qtree: name of the qtree
        :type qtree: str
        :param share_name: name of the share
        :type share_name: str
        :param group_name: force group name if provided (supported by cluster-mode filers with ontapi >= 1.30)
        :type group_name: str
        :param comment: description of the share
        :type comment: str
        :param umask: file permission umask
        :type umask: str
        :param vscan_fileop_profile: vscan-fileop-profile virus scan option (no_scan, standard, strict, writes_only)
        :type vscan_fileop_profile: str
        :raises NidhoggException: if an error occurs
        """
        opts = dict(
            dir_umask=umask,
            file_umask=umask,
            path="/vol/{0}/{1}".format(volume, qtree),
            share_name=share_name,
            vscan_fileop_profile=vscan_fileop_profile
        )
        if group_name and self.has_forcegroup:
            opts['force_group_for_create'] = group_name
        if comment:
            opts['comment'] = comment
        self.cifs_share_create(**opts)

    def set_cifs_acl(self, share_name, user="everyone", right=ACL_READ, set_group_rights=None):
        """Set a single ACL for the specifed share.

        :param share_name: name of the share
        :type share_name: str
        :param user: name of a user or group
        :type user: str
        :param right: right to be set, value must be one of :py:const:`~ACL_PERMISSIONS`
        :type right: str
        :param set_group_rights: if true, *user* param specifies a unix group name; if false, *user*
            param specifies a unix user name; if not defined, *user* param specifies a windows name
        :type set_group_rights: bool
        :raises NidhoggException: if an error occurs
        :raises NidhoggException: if wrong right was set
        """
        # check permissions
        if right not in self.ACL_PERMISSIONS:
            raise NidhoggException("Permission {0} not in {1}.".format(right, self.ACL_PERMISSIONS))

        # set user_group_type
        if set_group_rights is None:
            user_group_type = "windows"
        elif set_group_rights is True:
            user_group_type = "unix_group"
        else:
            user_group_type = "unix_user"
        opts = dict(
            permission=right,
            share=share_name,
            user_or_group=user,
            user_group_type=user_group_type
        )
        self.cifs_share_access_control_create(**opts)

    def list_cifs_acls(self, share_name, max_records=MAX_RECORDS):
        """Return ACL of the specified share.

        :param share_name: name of the share
        :type share_name: str
        :param max_records: limit returned records
        :type max_records: int
        :return: list of ACEs (access control entries)
        :rtype: :class:`~nidhogg.compatible.ACE` or empty list
        :raises NidhoggException: if an error occurs
        """
        opts = dict(
            query=dict(
                cifs_share_access_control=dict(
                    share=share_name
                )
            ),
            max_records=max_records
        )
        results = self.cifs_share_access_control_get_iter(**opts)["netapp"]["results"]
        if int(results["num-records"]) > 1:
            return [
                self._item_to_ace(item)
                for item in results['attributes-list']['cifs-share-access-control']
            ]
        elif int(results["num-records"]) == 1:
            return [
                self._item_to_ace(results['attributes-list']['cifs-share-access-control'])
            ]
        logger.warning("get_cifs_acl: no acls found")
        return []

    def delete_cifs_acl(self, share_name, user_or_group, is_group=None):
        """Delete cifs ACL of the specified user or group.

        :param share_name: name of the share
        :type share_name: str
        :param user_or_group: name of a user or group
        :type user_or_group: str
        :param is_group: not used for cluster-mode filers, specified here
            to be compatible with seven-mode method signature
        :type is_group: None
        :raises NidhoggException: if an error occurs
        """
        opts = dict(
            share=share_name,
            user_or_group=user_or_group
        )
        self.cifs_share_access_control_delete(**opts)

    def delete_cifs_acls(self, share_name):
        """Remove all cifs permssions.

        :param share_name: name of the share
        :type share_name: str
        :raises NidhoggException: if an error occurs
        """
        acls = self.list_cifs_acls(share_name)
        for ace in acls:
            self.delete_cifs_acl(
                share_name=ace["share_name"],
                user_or_group=ace["user_or_group"]
            )

    def set_quota(self, volume, qtree, quota_in_mb=1024, wait_til_finished=True):
        """Set a quota in MiB (default = 1GiB) for the specified volume and qtree.

        :param volume: name of the volume
        :type volume: str
        :param qtree: name of the qtree
        :type qtree: str
        :param quota_in_mb: quota in MiB
        :type quota_in_mb: int
        :param wait_til_finished: if false, do not wait for resize operation
        :type wait_til_finished: bool
        :raises NidhoggException: if an error occurs
        :raises NidhoggException: if resize did not finish in time and we were waiting for it
        :raises NidhoggException: if quotas are not enabled
        """
        quota_target = "/vol/{0}/{1}".format(volume, qtree)
        quota_type = "tree"
        quota_in_kb = int(round(quota_in_mb * 1024))
        # policy "default" must be specified for cluster-mode filers
        self.quota_set_entry(
            volume=volume,
            qtree="",
            disk_limit=quota_in_kb,
            soft_disk_limit=int(round(quota_in_kb * 0.8)),  # use 80% of the given quota as warn-limit
            quota_target=quota_target,
            quota_type=quota_type,
            policy="default"
        )
        self.quota_resize(volume=volume)
        if wait_til_finished:
            for i in range(0, nidhogg.core.QUOTA_RESIZE_WAIT_CYCLES):
                sleep(nidhogg.core.QUOTA_RESIZE_WAIT_TIME)
                status = self.quota_status(volume=volume)["netapp"]["results"]["status"]
                if status.lower() == "on":
                    return
                # check if quotas are turned on at all
                if status.lower() == "off":
                    raise NidhoggException("Quotas are not enabled.")
            # waiting for quote resize exceeded
            logger.debug("resize of {0}:/vol/{1} after setting quota for {2} did not finish".format(
                self.vserver_fqdn,
                volume,
                qtree
            ))
            raise NidhoggException("Quota resize did not finish in time.")

    def delete_quota(self, volume, qtree):
        """Delete the quota of the specified volume and qtree.

        :param volume: name of the volume
        :type volume: str
        :param qtree: name of the qtree
        :type qtree: str
        :raises NidhoggException: if an error occurs
        """
        quota_target = "/vol/{0}/{1}".format(volume, qtree)
        quota_type = "tree"
        # policy "default" must be specified for cluster-mode filers
        self.quota_delete_entry(
            volume=volume,
            qtree="",
            quota_target=quota_target,
            quota_type=quota_type,
            policy="default"
        )

    def update_snapmirror(self, volume):
        """Trigger the snapmirror replication. You have to be connected on the destination server.

        :param volume: name of snapmirror destination volume
        :type volume: str
        :raises NidhoggException: if an error occurs
        """
        self.snapmirror_update(
            destination_location="{}:{}".format(self.vserver, volume)
        )

    def update_snapmirror_with_snapshot(self, name, volume):
        """Trigger the snapmirror replication. You have to be connected on the destination server.

        :param name: name of the source snapshot
        :type name: str
        :param volume: name of snapmirror destination volume
        :type volume: str
        :raises NidhoggException: if an error occurs
        """
        self.snapmirror_update(
            destination_location="{}:{}".format(self.vserver, volume),
            source_snapshot=name
        )

    def get_snapmirror_status(self, volume=None, max_records=MAX_RECORDS):
        """Get status of snapmirror replication pairs. You have to be connected on the destination server.

        If no params are provided, return all snapmirror status pairs.

        :param volume: name of destination volume
        :type volume: str
        :return: list of all snapmirror pair status
        :rtype: list of :class:`~nidhogg.compatible.SnapmirrorStatus` or empty list
        :raises NidhoggException: if an error occurs
        """
        opts = dict()
        if volume:
            opts['query'] = dict(
                snapmirror_info=dict(
                    destination_location="{}:{}".format(self.vserver, volume)
                )
            )
            opts['max_records'] = max_records
        results = self.snapmirror_get_iter(**opts)["netapp"]["results"]
        if int(results["num-records"]) > 1:
            return [
                self._item_to_snapmirrorstatus(item)
                for item in results["attributes-list"]["snapmirror-info"]
            ]
        elif int(results["num-records"]) == 1:
            return [
                self._item_to_snapmirrorstatus(results["attributes-list"]["snapmirror-info"])
            ]
        logger.warn("get_snapmirror_status: no entries found")
        return []

    def get_snapmirror_volume_status(self, *args, **kwargs):
        """Not available for cluster mode."""
        raise NotImplementedError()     # pragma: no cover

    def create_snapshot(self, volume, name, label=None):
        """Create a snapshot with an optional label.

        :param volume: name of the volume
        :type volume: str
        :param name: name of the snapshot
        :type name: str
        :param label: add a snapmirror label to snapshot
        :type label: str
        :raises NidhoggException: if an error occurs
        """
        opts = dict()
        if label:
            opts['snapmirror_label'] = label
        opts['volume'] = volume
        opts['snapshot'] = name
        self.snapshot_create(**opts)

    def list_snapmirror_destinations(self, volume=None, max_records=MAX_RECORDS):
        """List all snapmirror destinations. You have to be connected on the source server.

        If no params are provided, return all snapmirror destinations.

        :param volume: name of source volume
        :type volume: str
        :return: list of all snapmirror destinations
        :rtype: list of :class:`~nidhogg.compatible.SnapmirrorDestinationInfo` or empty list
        :raises NidhoggException: if an error occurs
        """
        opts = dict()
        if volume:
            opts['query'] = dict(
                snapmirror_destination_info=dict(
                    source_location="{}:{}".format(self.vserver, volume)
                )
            )
            opts['max_records'] = max_records
        results = self.snapmirror_get_destination_iter(**opts)["netapp"]["results"]
        if int(results["num-records"]) > 1:
            return [
                self._item_to_snapmirrordestinationinfo(item)
                for item in results["attributes-list"]["snapmirror-destination-info"]
            ]
        elif int(results["num-records"]) == 1:
            return [
                self._item_to_snapmirrordestinationinfo(results["attributes-list"]["snapmirror-destination-info"])
            ]
        logger.warn("list_snapmirror_destinations: no entries found")
        return []



#!/usr/bin/env python2.7
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import os
from setuptools import setup


def read(fname):
    """Utility function to read the README file.

    Used for the long_description.  It's nice, because now 1) we have a top level
    README file and 2) it's easier to type in the README file than to put a raw
    string in below ...
    """
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="nidhogg",
    version="3.8.0",
    description="wrapper interface to Netapp filers",
    long_description=read("README.rst"),
    author="Christian Assing, Roland Wohlfahrt",
    author_email="christian.assing@infineon.com, roland.wohlfahrt@infineon.com",
    url="https://github.com/ifxit/nidhogg",
    packages=["nidhogg"],
    install_requires=open("requirements.txt").readlines(),
    license="MIT License, Copyright (c) 2018 Infineon Technologies AG",
    platforms="any",
    keywords=["netapp", "vserver", "sevenmode", "ontapi"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Natural Language :: English",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Archiving",
        "Topic :: System :: Systems Administration",
    ],
)
